//     File: Ind.cc
//  Version: 1.2, for BeIDE 1.3.1+
//   Author: Bjorn Tornqvist, Polyfemos @ IRC
//           pt96bto@student.hk-r.se
//           http://www.manda.rsn.hk-r.se/~bjorn/
//  Created: 23 March 1997 (dr8)
// Last Mod: 9 Oct 1997 (PR1)
//
// This is the source for the source code indenter add-on to BeIDE.
// It indents C, C++ and Java code.
// Note: It doesn't change the code style!
//
// Geek note: I've constructed the engine to make 2 passes through the
// source code. On the first pass, all lines are stripped from leading
// and following whitespaces and stored in a double-linked list (for
// easier traversing later on). Some recognition of what type of line
// it is is also done here, aswell as some preparation for the second
// pass. The second pass uses the information collected in the first
// pass and just indents. It's in this pass all logic is, and cosmetical
// errors should be sought out here.
// The first pass consists mainly of Line::Line() and perform_edit()
// the second pass is in processIndention() (all the logic is here)
// and createIndention() (which just creates the new nicely indented
// code block so that BeIDE can get it).
//
// As always, should you or your CW force be caught or killed reading
// this very ugly and utterly poorly documented source code, the
// secretary will disavow any knowledge of you and your actions.
// This disclaimer will self destruct in 5 seconds.

#include "Ind.hh"
#include "Stack.h"
#include <string.h>
#include <stdio.h>
#include <ctype.h>
#include "MTextAddOn.h"

bool equals(const char*, const char*, int maxCheckLen=5);

#pragma export on
extern "C" {
	long perform_edit(MTextAddOn*);
}
#pragma export reset

bool
equals(const char* two, const char* one, int maxCheckLen)
{
	const char* it1=one;
	const char* it2=two;
	while(*it1==*it2)
	{
		it1++;
		it2++;
	}
	if((*it2==0)&&(maxCheckLen==4))
		return true;
	if((*it2==0)&&((maxCheckLen==5)&&(*it1!=' ')))
		return true;
	return false;
}

Line::Line(const char* theLine, long maxLen)
{
	//copy the data to the internal buffer
	int len=0;
	const char* tmp=theLine;
	for(len=0;((*(tmp+len))&&(*(tmp+len)!='\n')&&(len<maxLen));) len++;
	myRow=new char [len+2];
	strncpy(myRow, theLine, len);
	myRow[len]=0;
	//process the text
	const char* iterator=theLine;
	bool seenCComment=false;
	bool seenCCComment=false;
	bool charWarn=false;
	myStrangeCorrected=false;
	myCurlySum=0;
	myParantesisSum=0;
	myLength=0;
	myIndention=0;
	myLastChar=0;
	myFirstChar=*iterator;
	myIsStartOfCComment=false;
	myIsEndOfCComment=false;
	myIsCCComment=false;
	myPreviousLine=0;
	myNextLine=0;
	
	while(*iterator!='\n')
	{
		// Check for a C++ comment
		if((*iterator=='/')&&(*(iterator+1)=='/')&&myIsCCComment==false)
		{
			if(myLength==0)
				myIsCCComment=true;
			else
			{
				myLastChar=myPreviousChar(iterator-1);
				while(*iterator!='\n')
				{
					myLength++;
					iterator++;
				}	
				break;
			}
		}
		//Check for start of a C comment
		if((*iterator=='/')&&(*(iterator+1)=='*'))
		{
			if(myLength==0)
				myIsStartOfCComment=true;
			else
				myLastChar=myPreviousChar(iterator-1);
		}
		//Check for end of a C comment
		if((*iterator=='*')&&(*(iterator+1)=='/'))
		{
			if(myLength==0)
				myIsEndOfCComment=true;
		}
		
		//the indention.
		switch (*iterator)
		{
		case '(':
		{
			myParantesisSum++;
		}
			break;
		case ')':
		{
			myParantesisSum--;
		}
			break;
		case '{': {
				if(!charWarn)
					myCurlySum++;
			}
			break;
		case '}':
			if(!charWarn)
				myCurlySum--;
			break;
		case '\'':
			charWarn=(charWarn) ? false : true;
		}
		myLength++;
		iterator++;
	}
	if(!myLastChar)
		myLastChar=*(iterator-1);
}

Line::~Line()
{
	delete [] myRow;
}

void
Line::setIndention(int theInd)
{
	myIndention=theInd;
}

int
Line::getLength()
{
	return myLength;
}

char
Line::myPreviousChar(const char* theOffset)
{
	const char* iterator=theOffset;
	while((*iterator==' ')||
		(*iterator=='\t'))
		iterator--;
	return (*iterator);
}

Indent::Indent()
{
	myFirstLine=0;
	myLastLine=0;
	myLen=0;
}

Indent::~Indent()
{
	Line* iterator;
	iterator=myFirstLine;
	if(myFirstLine==0) return;
	while(iterator!=0)
	{
		Line* thisLine=iterator;
		iterator=iterator->myNextLine;
		delete thisLine;
	}
}

void
Indent::addLine(const char* theLine, long maxLen)
{
	if(myFirstLine)
	{
		myLastLine->myNextLine=new Line(theLine, maxLen);
		myLen+=myLastLine->myNextLine->getLength();
		myLastLine->myNextLine->myPreviousLine=myLastLine;
		myLastLine=myLastLine->myNextLine;
	}
	else
	{
		myFirstLine=new Line(theLine, maxLen);
		myLen+=myFirstLine->getLength();
		myFirstLine->myNextLine=0;
		myLastLine=myFirstLine;
	}
}

void
Indent::processIndention(int theStartIndention)
{
	if(!myFirstLine) return;
	Line* iterator=myFirstLine->myNextLine;
	Stack stack;
	stack.push(theStartIndention);
	int emergency;
	bool needsStrangeCorrection=false;
	bool caseSpecial=false;
	int strangeCorrectionDelta=0;
	while(iterator)
	{
		if((iterator->myIsCCComment==true))
		{
			iterator->myIndention=stack.sneak();
			if(caseSpecial) iterator->myIndention--;
			iterator=iterator->myNextLine;
			continue;
		}
		else if(iterator->myFirstChar=='#')
			iterator->myIndention=0;
		else if (iterator->myLastChar==':')
			iterator->myIndention=stack.sneak()-1;
		else if(iterator->myCurlySum!=0)
		{
			// adjust Indention on the next row if '{' encountered
			if(iterator->myCurlySum>0)
			{
				emergency=1;
				int pushIndention=iterator->myPreviousLine->myIndention;
				
				if(needsStrangeCorrection)
				{
					iterator->myIndention=strangeCorrectionDelta;
					pushIndention=strangeCorrectionDelta;
					needsStrangeCorrection=false;
					strangeCorrectionDelta=0;
				}
				else
					iterator->myIndention=iterator->myPreviousLine->myIndention;
				if(equals("case", iterator->myPreviousLine->myRow, 4)&&(iterator->myPreviousLine->myLastChar!='{'))
				{
					pushIndention++;
					iterator->myIndention++;
				}
				if(equals("case", iterator->myRow, 4)&&(iterator->myLastChar=='{'))
				{
					pushIndention++;
					iterator->myIndention--;
					caseSpecial=true;
				}
				if(equals("else", iterator->myRow, 4))
				{
					iterator->myIndention--;
					pushIndention--;
				}
				if((iterator->myPreviousLine->myLastChar==')')&&
					(iterator->myLastChar=='{')&&
					(iterator->myFirstChar!='{'))
				{
					iterator->myIndention++;
					emergency=2;
				}
				int num=iterator->myCurlySum;
				while(num-->0)
					stack.push(pushIndention+emergency);
			}
			if(iterator->myCurlySum<0)
			{
				int num=iterator->myCurlySum;
				iterator->myIndention=stack.pop()-1;
				num++;
				while(num<0)
				{
					iterator->myIndention=stack.pop()-1;
					num++;
				}
				if(caseSpecial) iterator->myIndention--;
				caseSpecial=false;
				if(iterator->myIndention<0)
					iterator->myIndention=iterator->myPreviousLine->myIndention-1;
				needsStrangeCorrection=false;
				strangeCorrectionDelta=0;
			}
		}
		else
		{
			iterator->myIndention=stack.sneak();
			if(caseSpecial) iterator->myIndention--;
			if(iterator->myIndention<0)
				iterator->myIndention=iterator->myPreviousLine->myIndention;
			if((iterator->myPreviousLine->myLastChar==')')||
				(equals("else", iterator->myPreviousLine->myRow)==true))
			{
				iterator->myIndention=(iterator->myPreviousLine->myStrangeCorrected==true)?
					strangeCorrectionDelta+1:(iterator->myPreviousLine->myIsCCComment)?
					iterator->myPreviousLine->myIndention:iterator->myPreviousLine->myIndention+1;
				strangeCorrectionDelta=0;
				needsStrangeCorrection=false;
			}
			else if((iterator->myPreviousLine->myLastChar==',')||
				(iterator->myPreviousLine->myLastChar=='(')||
				((iterator->myPreviousLine->myRow[iterator->myPreviousLine->myLength-1]=='&')&&
				(iterator->myPreviousLine->myRow[iterator->myPreviousLine->myLength-2]=='&'))||
				((iterator->myPreviousLine->myRow[iterator->myPreviousLine->myLength-1]=='/')&&
				(iterator->myPreviousLine->myRow[iterator->myPreviousLine->myLength-2]!='*'))||
				(iterator->myPreviousLine->myLastChar=='|')||
				(iterator->myPreviousLine->myLastChar=='+')||
				(iterator->myPreviousLine->myLastChar=='?')||
				(iterator->myPreviousLine->myLastChar=='-'))
			{
				if(iterator->myPreviousLine->myStrangeCorrected)
					iterator->myIndention=iterator->myPreviousLine->myIndention;
				else
					iterator->myIndention=iterator->myPreviousLine->myIndention+1;
				if(iterator->myPreviousLine->myIsCCComment) iterator->myIndention=iterator->myPreviousLine->myIndention;
				if(!needsStrangeCorrection)
					strangeCorrectionDelta=iterator->myPreviousLine->myIndention;
				needsStrangeCorrection=true;
				iterator->myStrangeCorrected=true;
			}
		}
		iterator=iterator->myNextLine;
	}
}

//This method just inserts the right amount of tabs into the output.
void
Indent::generateIndention(char* output,
	bool stripComments,
	bool stripIndention)
{
	if(output==0)
		return;
	char* oiter=output;
	char* c;
	Line* iterator=myFirstLine;
	while(iterator)
	{
		int ind=iterator->myIndention;
		if((ind>0)&&(!stripIndention))
			while(ind--)
				*oiter++='\t';
		c=iterator->myRow;
		int tempLen=0;
		while((*c!='\n')&&(tempLen++<iterator->myLength))
			*oiter++=*c++;
		*oiter++='\n';
		iterator=iterator->myNextLine;
		if(stripComments&&(iterator!=0))
			if(iterator->myIsCCComment==true)
				if(iterator->myLastChar==';')
					iterator=iterator->myNextLine;
	}
	*oiter=0;
}

long
perform_edit(
	MTextAddOn* theAddon)
{
	bool shift=!!(modifiers() & B_SHIFT_KEY);
	bool ctrl=!!(modifiers() & B_COMMAND_KEY);
	
	long myStartOffset, myEndOffset;
	
	const char* myIterator;
	char* myOldText;
	const char* myText=theAddon->Text();
	char* myOldTextIterator;
	char* myNewText;
	char* myNewTextIterator;
	int currentIndention=0;
	try
	{
		theAddon->GetSelection(&myStartOffset, &myEndOffset);
		if (myEndOffset <= myStartOffset)
			return (B_ERROR);
		long totalLength=myEndOffset-myStartOffset;
		myOldText=new char [myEndOffset-myStartOffset+10];
		myOldTextIterator=myOldText;
		// remove all tabs in the selection.
		int myNumOfLines=0;
		int startOfLine=false;
		bool beenHere=0;
		bool doneThat=0;
		myIterator=myText+myStartOffset;
		// Since the selection can start halways into the leading whitespaces
		// of the first line, I have to look for the start-indentation on the
		// second line.
		while(*myIterator!='\n')
		{
			myIterator++;
			if(myIterator==(myText+myEndOffset))
			{
				delete [] myOldText;
				beep();
				return B_NO_ERROR;
			}
		}
		myIterator++;
		if(myIterator==(myText+myEndOffset))
		{
			delete [] myOldText;
			beep();
			return B_NO_ERROR;
		}
		int indentationSpaces=0;
		while((*myIterator=='\t')||(*myIterator==' '))
		{
			if(*myIterator==' ') indentationSpaces++;
			else currentIndention++;
			myIterator++;
			if(myIterator==(myText+myEndOffset))
			{
				delete [] myOldText;
				beep();
				return B_NO_ERROR;
			}
		}
		currentIndention+=(int)indentationSpaces/3;
		myIterator=myText+myStartOffset;
		int counter=0;
		do
		{
			if((startOfLine)&&((*myIterator=='\t')||
				(*myIterator==' ')))
			{
				myIterator++;
				totalLength--;
			}
			else
			{
				startOfLine=false;
				*(myOldTextIterator++)=(*myIterator);
				if(*myIterator=='\n'){
					startOfLine=true;
					myNumOfLines++;
					//erase whitespaces to the right of the line
					myOldTextIterator--;
					while(((*(myOldTextIterator-1)=='\t')||
						(*(myOldTextIterator-1)==' '))&&
						(myOldTextIterator!=myOldText))
						myOldTextIterator--;
					*myOldTextIterator++='\n';
				}
				myIterator++;
			}
		} while((myIterator<(myText+myEndOffset))&&(*myIterator!=0)&&(counter++<(myEndOffset-myStartOffset)));
		
		*myOldTextIterator=0;
		Indent* indent=new Indent();
		indent->addLine(myOldText, totalLength);
		char* hej=myOldText;
		int bytes=0;
		for(int index=0;(index<myNumOfLines)&&(totalLength>0);index++)
		{
			while((*hej)!='\n')
			{
				hej++;
				totalLength--;
				if(hej>=myOldText+myEndOffset-myStartOffset)
				{
					delete []myOldText;
					delete indent;
					beep();
					return B_ERROR;
				}
			}
			hej++;
			totalLength--;
			indent->addLine(hej, totalLength);
		}
		
		indent->processIndention(currentIndention);
		int len=indent->getLength();
		char* result=new char [len];
		indent->generateIndention(result, ctrl, shift);
		if(result!=0)
		{
			theAddon->Delete();
			theAddon->Insert(result);
			delete [] result;
		}
		delete [] myOldText;
		delete indent;
		return B_NO_ERROR;
	}
	catch(...)
	{
		beep();
		return B_ERROR;
	}
}
